

#include <stdio.h>
#include <windows.h>
#include "squish.h"
#include "colourblock.h"
#include <cstring>
#include "blpjpeg.h"
#include "dxt.h"
#include "resource.h"
#include "commctrl.h"
#include "AFilemic.h"

HINSTANCE  hDLLModule;
BOOL APIENTRY DllMain(HINSTANCE  hModule, DWORD  ul_reason_for_call, LPVOID lpReserved)
{
    switch (ul_reason_for_call) {
    case DLL_PROCESS_ATTACH:
    case DLL_PROCESS_DETACH:
    case DLL_THREAD_ATTACH:
    case DLL_THREAD_DETACH:
        break;
    }
    hDLLModule = hModule;
    return TRUE; 
}

#define API __stdcall

#define GFP_RGB	0
#define GFP_BGR	1

#define GFP_READ	0x0001
#define GFP_WRITE 0x0002

typedef struct {
    unsigned char red[256];
    unsigned char green[256];
    unsigned char blue[256];
} GFP_COLORMAP;


#define FILETYPE_BLP "BLP Blizzard Image format"
#define EXTENSION_BLP "blp"
#define DXTCOMPRESSION "DXT compression"
#define BLP0 0x30504c42
#define BLP1 0x31504c42
#define BLP2 0x32504c42
#define BLP1ALPHA 8
#define PALETTED 1
#define DXTCOMPRESSED 2
#define UNCOMPRESSED 3
#define DXT1 0
#define DXT3 0x10000
#define DXT5 0x70000
typedef struct BLPHeader {
    DWORD Signature;
    DWORD PixelCompression;
    DWORD Flags;
    DWORD Width;
    DWORD Height;
    union {
        struct {
            DWORD alphaEncoding;
            DWORD flags2;
            DWORD offsets[16];
            DWORD lengths[16];
        }blp1;
        struct {
            DWORD offsets[16];
            DWORD lengths[16];
        }blp2;
    }HeaderEnd;
} BLPHeader;
typedef struct {
    //There are these three fields in SDK. But host doesn't use them
    //FILE * fp;
    //INT width;
    //INT component_per_pixel;
    HANDLE fp;
    INT width;
    INT component_per_pixel;
    INT bits_per_pixel;
    BLPHeader* header;
    DWORD firstMipmap;
    int type;
    BYTE* filePalette;
    BYTE* buf;
    DWORD curLine;
 } MYDATA;
typedef unsigned long long int QWORD;
static float rgbtoxyz[256];
DWORD isinitialised = 0;
void initLab()
{
    if (isinitialised)return;
    for (DWORD i = 256; i--;) {
        float a = (float)i / 255.0f;
        if (a > 0.04045f)a = powf(((a + 0.055f) / 1.055f), 2.4f);
        else a = a *0.0774f;
        rgbtoxyz[i] = a;
    }
    isinitialised++;
}

static float delta76(DWORD R1, DWORD G1, DWORD B1, DWORD R2, DWORD G2, DWORD B2)
{
    float x1, y1, z1, r, g, b, x2, y2, z2;
    r = rgbtoxyz[R1];
    g = rgbtoxyz[G1];
    b = rgbtoxyz[B1];


    x1 = 0.4339528f*r + 0.3762194f *g + 0.1898278f*b;
    y1 = 0.212671f*r + 0.71516f *g + 0.072169f*b;
    z1 = 0.0177579f*r + 0.1094765f*g + 0.8727656f*b;
    if (x1 <= 8.856452e-3f)x1 = 7.787037f*x1 + 0.137931f;
    else if (x1 < 0.015625f) { x1 *= 64.0f; x1 = 0.25f*((x1 * 0.24379f + 0.958f)*x1 + 0.07892f) / (x1 + 0.27962f); }
    else if (x1 <0.125f) { x1 *= 8.0f; x1 = 0.5f*((x1 * 0.24379f + 0.958f)*x1 + 0.07892f) / (x1 + 0.27962f); }
    else x1 = ((x1 * 0.24379f + 0.958f)*x1 + 0.07892f) / (x1 + 0.27962f);
    if (y1 <= 8.856452e-3f)y1 = 7.787037f*y1 + 0.137931f;
    else if (y1 < 0.015625f) { y1 *= 64.0f; y1 = 0.25f*((y1 * 0.24379f + 0.958f)*y1 + 0.07892f) / (y1 + 0.27962f); }
    else if (y1 <0.125f) { y1 *= 8.0f; y1 = 0.5f*((y1 * 0.24379f + 0.958f)*y1 + 0.07892f) / (y1 + 0.27962f); }
    else y1 = ((y1 * 0.24379f + 0.958f)*y1 + 0.07892f) / (y1 + 0.27962f);
    if (z1 <= 8.856452e-3f)z1 = 7.787037f*z1 + 0.137931f;
    else if (z1 < 0.015625f) { z1 *= 64.0f; z1 = 0.25f*((z1 * 0.24379f + 0.958f)*z1 + 0.07892f) / (z1 + 0.27962f); }
    else if (z1<0.125f) { z1 *= 8.0f; z1 = 0.5f*((z1 * 0.24379f + 0.958f)*z1 + 0.07892f) / (z1 + 0.27962f); }
    else z1 = ((z1 * 0.24379f + 0.958f)*z1 + 0.07892f) / (z1 + 0.27962f);

    r = rgbtoxyz[R2];
    g = rgbtoxyz[G2];
    b = rgbtoxyz[B2];

    x2 = 0.4339528f*r + 0.3762194f *g + 0.1898278f*b;
    y2 = 0.212671f*r + 0.71516f *g + 0.072169f*b;
    z2 = 0.0177579f*r + 0.1094765f*g + 0.8727656f*b;
    if (x2 <= 8.856452e-3f)x2 = 7.787037f*x2 + 0.137931f;
    else if (x2 < 0.015625f) { x2 *= 64.0f; x2 = 0.25f*((x2 * 0.24379f + 0.958f)*x2 + 0.07892f) / (x2 + 0.27962f); }
    else if (x2<0.125f) { x2 *= 8.0f; x2 = 0.5f*((x2 * 0.24379f + 0.958f)*x2 + 0.07892f) / (x2 + 0.27962f); }
    else x2 = ((x2 * 0.24379f + 0.958f)*x2 + 0.07892f) / (x2 + 0.27962f);
    if (y2 <= 8.856452e-3f)y2 = 7.787037f*y2 + 0.137931f;
    else if (y2 < 0.015625f) { y2 *= 64.0f; y2 = 0.25f*((y2 * 0.24379f + 0.958f)*y2 + 0.07892f) / (y2 + 0.27962f); }
    else if (y2 <0.125f) { y2 *= 8.0f; y2 = 0.5f*((y2 * 0.24379f + 0.958f)*y2 + 0.07892f) / (y2 + 0.27962f); }
    else y2 = ((y2 * 0.24379f + 0.958f)*y2 + 0.07892f) / (y2 + 0.27962f);
    if (z2 <= 8.856452e-3f)z2 = 7.787037f*z2 + 0.137931f;
    else if (z2 < 0.015625f) { z2 *= 64.0f; z2 = 0.25f*((z2 * 0.24379f + 0.958f)*z2 + 0.07892f) / (z2 + 0.27962f); }
    else if (z2 <0.125f) { z2 *= 8.0f; z2 = 0.5f*((z2 * 0.24379f + 0.958f)*z2 + 0.07892f) / (z2 + 0.27962f); }
    else z2 = ((z2 * 0.24379f + 0.958f)*z2 + 0.07892f) / (z2 + 0.27962f);

    r = 116.0f * (y1 - y2);
    g = 500.0f * (x1 - y1 - x2 + y2);
    b = 200.0f * (y1 - z1 - y2 + z2);
    return sqrtf(r*r + g*g + b*b);
}

//
//static float delta2000(DWORD R1, DWORD G1, DWORD B1, DWORD R2, DWORD G2, DWORD B2)
//{
//    float L1, a1, b1, L2, a2, b2, x, y, z;
//    L1 = rgbtoxyz[R1];
//    a1 = rgbtoxyz[G1];
//    b1 = rgbtoxyz[B1];
//    x = 0.4339528f*L1 + 0.3762194f *a1 + 0.1898278f*b1;
//    y = 0.212671f*L1 + 0.71516f *a1 + 0.072169f*b1;
//    z = 0.0177579f*L1 + 0.1094765f*a1 + 0.8727656f*b1;
//    if (x <= 8.856452e-3f)x = 7.787037f*x + 0.137931f;
//    else x = cbrtf(x);
//    if (y <= 8.856452e-3f)y = 7.787037f*y + 0.137931f;
//    else y = cbrtf(y);
//    if (z <= 8.856452e-3f)z = 7.787037f*z + 0.137931f;
//    else z = cbrtf(z);
//    L1 = 116.0f * y - 16.0f;
//    a1 = 500.0f * (x - y);
//    b1 = 200.0f * (y - z);
//
//    L2 = rgbtoxyz[R2];
//    a2 = rgbtoxyz[G2];
//    b2 = rgbtoxyz[B2];
//    x = 0.4339528f*L2 + 0.3762194f *a2 + 0.1898278f*b2;
//    y = 0.212671f*L2 + 0.71516f *a2 + 0.072169f*b2;
//    z = 0.0177579f*L2 + 0.1094765f*a2 + 0.8727656f*b2;
//    if (x <= 8.856452e-3f)x = 7.787037f*x + 0.137931f;
//    else x = cbrtf(x);
//    if (y <= 8.856452e-3f)y = 7.787037f*y + 0.137931f;
//    else y = cbrtf(y);
//    if (z <= 8.856452e-3f)z = 7.787037f*z + 0.137931f;
//    else z = cbrtf(z);
//    L2 = 116.0f * y - 16.0f;
//    a2 = 500.0f * (x - y);
//    b2 = 200.0f * (y - z);
//
//    double Ca = (sqrtf(a1*a1 + b1*b1) + sqrtf(a2*a2 + b2*b2))*0.5;
//    double Ca1 = Ca *Ca*Ca*Ca*Ca*Ca*Ca;
//    double G = (1.0 - sqrt(Ca1 / (6103515625.0 + Ca1)))*0.5;
//    double a1_ = (1.0 + G)*a1;
//    double a2_ = (1.0 + G)*a2;
//    double C1 = sqrt(a1_*a1_ + b1* b1);
//    double C2 = sqrt(a2_*a2_ + b2* b2);
//    double dC = C2 - C1;
//    double h1, h2, dh;
//    if (b1 >= 0.0)h1 = 57.29578*atan2(b1, a1_);
//    else h1 = 57.29578*atan2(b1, a1_) + 360.0;
//    if (b2 >= 0.0)h2 = 57.29578*atan2(b2, a2_);
//    else h2 = 57.29578*atan2(b2, a2_) + 360.0;
//    if ((h2 - h1) <= 180.0 && (h2 - h1) >= -180.0)dh = h2 - h1;
//    else if ((h2 - h1)>180.0)dh = h2 - h1 - 360.0;
//    else dh = h2 - h1 + 360.0;
//
//    double dL = L2 - L1;
//    double dH = 2.0*sqrt(C1*C2)*sin(dh * 0.008726646);
//    double Lave = (L1 + L2) *0.5;
//    double Cave = (C1 + C2) *0.5;
//    double have;
//    if (C1*C2 == 0.0)have = h1 + h2;
//    else if (fabs(h2 - h1)<180.0)have = (h1 + h2) *0.5;
//    else if ((h1 + h2)<360.0)have = (h1 + h2) *0.5 + 180.0;
//    else have = (h1 + h2) *0.5 - 180.0;
//    double Lave2 = Lave - 50.0; Lave2 *= Lave2;
//    double SL = 1.0 + 0.015*Lave2 / sqrt(20.0 + Lave2);
//    double SC = 1.0 + 0.045*Cave;
//    double T = 1.0 - 0.17*cos(0.01745329*(have - 30.0)) + 0.24*cos(0.01745329*(have + have)) + 0.32*cos(0.01745329*(3.0*have + 6.0)) - 0.2*cos(0.01745329*(4.0*have - 63.0));
//    double SH = 1.0 + 0.015*Cave*T;
//    have = (have - 275.0) *0.04;
//    double dTheta = 30.0*exp(-1.0*have*have);
//    double RCt = Cave*Cave*Cave*Cave*Cave*Cave*Cave;
//    double RC = 2.0*sqrt(RCt / (RCt + 6103515625.0));
//    double RT = -sin(0.01745329*dTheta)*RC;
//    double AJ = dL / SL;
//    double AK = dC / SC;
//    double AL = dH / SH;
//    return (float)sqrt(AJ*AJ + AK*AK + AL*AL + AK*AL*RT);
//}

typedef struct piped {
    BYTE redMin;
    BYTE redMax;
    BYTE greenMin;
    BYTE greenMax;
    BYTE blueMin;
    BYTE blueMax;
    BYTE maxSide;
    BYTE diametr;
    DWORD pxCount;
    union {
        DWORD ra;
        DWORD count;
    };
    union {
        DWORD ga;
        BYTE minmax[4];
    };

    union {
        DWORD ba;
        float maxErr;
    };
    /*void calcmaxErr2()
    {
    DWORD mr, mg, mb;
    float ft;
    mr = (redMax + redMin) >> 1;
    mg = (greenMax + greenMin) >> 1;
    mb = (blueMax + blueMin) >> 1;
    if (redMin == redMax)maxErr = 0.0f;
    else maxErr = delta2000(redMin, mg, mb, redMax, mg, mb);

    maxSide = 0;
    if (greenMin == greenMax)ft = 0.0f;
    else ft = delta2000(mr, greenMin, mb, mr, greenMax, mb);
    if (ft > maxErr) { maxErr = ft; maxSide = 1; }
    if (blueMin == blueMax)ft = 0.0f;
    else ft = delta2000(mr, mg, blueMin, mr, mg, blueMax);
    if (ft > maxErr) { maxErr = ft; maxSide = 2; }
    maxErr *= sqrtf((float)pxCount);
    }*/
    void calcmaxErr1()
    {
        DWORD mr, mg, mb;
        float ft;
        mr = (redMax + redMin) >> 1;
        mg = (greenMax + greenMin) >> 1;
        mb = (blueMax + blueMin) >> 1;
        if (redMin == redMax)maxErr = 0.0f;
        else maxErr = delta76(redMin, mg, mb, redMax, mg, mb);

        maxSide = 0;
        if (greenMin == greenMax)ft = 0.0f;
        else ft = delta76(mr, greenMin, mb, mr, greenMax, mb);
        if (ft > maxErr) { maxErr = ft; maxSide = 1; }
        if (blueMin == blueMax)ft = 0.0f;
        else ft = delta76(mr, mg, blueMin, mr, mg, blueMax);
        if (ft > maxErr) { maxErr = ft; maxSide = 2; }
        maxErr *= pxCount;
    }

}piped;

//#pragma pack()

static DWORD GetPaletteMCCQ(BYTE* buf, piped* pipeds, const BYTE* source, DWORD pixels, DWORD colorDepth)
{
    //Median-cut color quantization
    //source 24-bit or 32-bit colour
    //buf out palette pipeds - buffer size= 0x1800
    initLab();
    DWORD pipedCount = 1, ofs, a, a1;

    BYTE b, b1;
    //diametr maxc2
    //

    pipeds[0].pxCount = ofs = pixels;
    ofs *= colorDepth;
    pipeds[0].redMax = pipeds[0].redMin = *(source);
    pipeds[0].greenMax = pipeds[0].greenMin = *(source + 1);
    pipeds[0].blueMax = pipeds[0].blueMin = *(source + 2);
    //inicial crop
    for (; ofs -= colorDepth;) {
        b = *(source + ofs);
        if (pipeds[0].redMin>b)pipeds[0].redMin = b;
        else if (pipeds[0].redMax<b)pipeds[0].redMax = b;
        b = *(source + ofs + 1);
        if (pipeds[0].greenMin>b)pipeds[0].greenMin = b;
        else if (pipeds[0].greenMax<b)pipeds[0].greenMax = b;
        b = *(source + ofs + 2);
        if (pipeds[0].blueMin>b)pipeds[0].blueMin = b;
        else if (pipeds[0].blueMax<b)pipeds[0].blueMax = b;
    }
    pipeds[0].calcmaxErr1();

    for (; pipedCount<256;) {
        //chooze piped or breake
        DWORD c = 0, mc;
        BYTE b2;
        float ft = pipeds[0].maxErr;
        for (a1 = 1; a1<pipedCount; a1++)
            if (ft < pipeds[a1].maxErr) {
                c = a1;
                ft = pipeds[a1].maxErr;
            }
        if (ft == 0.0f)break;
        //zero pool
        for (ofs = 256, a = 0xff00ff; ofs--;) {
            pipeds[ofs].count = 0;
            pipeds[ofs].ga = a;
        }
        //put pool
        ofs = pixels*colorDepth;
        mc = pipeds[c].maxSide;
        do {
            ofs -= colorDepth;
            b = *(source + ofs);
            if (pipeds[c].redMax >= b&&pipeds[c].redMin <= b) {
                b1 = *(source + ofs + 1);
                if (pipeds[c].greenMax >= b1&&pipeds[c].greenMin <= b1) {
                    a = b;
                    b = *(source + ofs + 2);
                    if (pipeds[c].blueMax >= b&&pipeds[c].blueMin <= b) {
                        if (mc == 1) { b2 = b; b = b1; b1 = b2; }
                        else if (!mc) { b2 = b; b = (BYTE)a; a = b2; }
                        pipeds[b].ra++;
                        a1 = pipeds[b].ga;
                        if (pipeds[b].minmax[0]>(BYTE)a)pipeds[b].minmax[0] = (BYTE)a;
                        if (pipeds[b].minmax[1]<(BYTE)a)pipeds[b].minmax[1] = (BYTE)a;
                        if (pipeds[b].minmax[2]>b1)pipeds[b].minmax[2] = b1;
                        if (pipeds[b].minmax[3]<b1)pipeds[b].minmax[3] = b1;
                    }
                }
            }
        } while (ofs);
        //divide piped c
        a = 0; a1 = pipeds[c].pxCount;
        if (mc == 2) { b1 = pipeds[c].blueMin; b2 = pipeds[c].blueMax; }
        else if (mc) { b1 = pipeds[c].greenMin; b2 = pipeds[c].greenMax; }
        else { b1 = pipeds[c].redMin; b2 = pipeds[c].redMax; }
        for (b = b1; b<b2; b++) {
            ofs = pipeds[b].count;
            if (a1>a + ofs) { a1 -= ofs; a += ofs; }
            else {
                if ((a1 == a + ofs) && ofs) {
                    if (b2 - b>b - b1) { a1 -= ofs; a += ofs; }
                    else b--;
                }
                else b--;
                break;
            }
        }
        if (b == b2)b--;
        pipeds[c].pxCount = a1;
        pipeds[pipedCount].pxCount = a;
        for (ofs = b + 1; ofs < b2&&!pipeds[ofs].ra; ofs++);
        for (; b>b1&&!pipeds[b].ra; b--);
        if (mc == 2) {
            pipeds[c].blueMin = (BYTE)ofs;
            pipeds[pipedCount].blueMax = b;
            pipeds[pipedCount].blueMin = b1;
        }
        else if (mc) {
            pipeds[c].greenMin = (BYTE)ofs;
            pipeds[pipedCount].greenMax = b;
            pipeds[pipedCount].greenMin = b1;
        }
        else {
            pipeds[c].redMin = (BYTE)ofs;
            pipeds[pipedCount].redMax = b;
            pipeds[pipedCount].redMin = b1;
        }
        //crop
        a1 = 0xff00ff;
        for (DWORD b11 = b1; b11 <= b; b11++) {
            a = pipeds[b11].ga;
            if ((BYTE)a1 > (BYTE)a)a1 = (a1 & 0xffffff00) | (BYTE)a;
            if ((a1 & 0xff00)<(a & 0xff00))a1 = (a1 & 0xffff00ff) | (a & 0xff00);
            if ((a1 & 0xff0000)>(a & 0xff0000))a1 = (a1 & 0xff00ffff) | (a & 0xff0000);
            if ((a1 & 0xff000000)<(a & 0xff000000))a1 = (a1 & 0xffffff) | (a & 0xff000000);
        }
        if (mc == 2) {
            pipeds[pipedCount].redMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[pipedCount].redMax = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[pipedCount].greenMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[pipedCount].greenMax = (BYTE)a1;
        }
        else if (mc) {
            pipeds[pipedCount].redMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[pipedCount].redMax = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[pipedCount].blueMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[pipedCount].blueMax = (BYTE)a1;
        }
        else {
            pipeds[pipedCount].blueMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[pipedCount].blueMax = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[pipedCount].greenMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[pipedCount].greenMax = (BYTE)a1;
        }
        //crop
        a1 = 0xff00ff;
        for (; ofs <= b2; ofs++) {
            a = pipeds[ofs].ga;
            if ((BYTE)a1>(BYTE)a)a1 = (a1 & 0xffffff00) | (BYTE)a;
            if ((a1 & 0xff00)<(a & 0xff00))a1 = (a1 & 0xffff00ff) | (a & 0xff00);
            if ((a1 & 0xff0000)>(a & 0xff0000))a1 = (a1 & 0xff00ffff) | (a & 0xff0000);
            if ((a1 & 0xff000000)<(a & 0xff000000))a1 = (a1 & 0xffffff) | (a & 0xff000000);
        }
        if (mc == 2) {
            pipeds[c].redMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[c].redMax = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[c].greenMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[c].greenMax = (BYTE)a1;
        }
        else if (mc) {
            pipeds[c].redMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[c].redMax = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[c].blueMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[c].blueMax = (BYTE)a1;
        }
        else {
            pipeds[c].blueMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[c].blueMax = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[c].greenMin = (BYTE)a1;
            a1 = a1 >> 8;
            pipeds[c].greenMax = (BYTE)a1;
        }

        pipeds[c].calcmaxErr1();
        pipeds[pipedCount].calcmaxErr1();

        pipedCount++;
    }
    //calculate middle colors
    for (ofs = 256, a = 0; ofs--;)pipeds[ofs].ra = pipeds[ofs].ga = pipeds[ofs].ba = a;
    ofs = pixels*colorDepth;
    for (; ofs -= colorDepth;) {
        b = *(source + ofs);
        b1 = *(source + ofs + 1);
        a = *(source + ofs + 2);
        for (a1 = 0; a1 < pipedCount; a1++) {
            if (pipeds[a1].redMax >= b&&pipeds[a1].redMin <= b) {
                if (pipeds[a1].greenMax >= b1&&pipeds[a1].greenMin <= b1) {
                    if (pipeds[a1].blueMax >= a&&pipeds[a1].blueMin <= a) {
                        pipeds[a1].ra += b;
                        pipeds[a1].ga += b1;
                        pipeds[a1].ba += a;
                        break;
                    }
                }
            }
        }
    }

    DWORD* bufd = (DWORD*)buf;
    for (ofs = 0; ofs < pipedCount; ofs++) {
        a1 = pipeds[ofs].pxCount >> 1;
        pipeds[ofs].ba += a1;
        pipeds[ofs].ga += a1;
        pipeds[ofs].ra += a1;
        a1 = pipeds[ofs].ba / pipeds[ofs].pxCount;
        a1 = a1 << 8;
        a = pipeds[ofs].ga / pipeds[ofs].pxCount;
        a1 += a;
        a1 = a1 << 8;
        a = pipeds[ofs].ra / pipeds[ofs].pxCount;
        a1 += a + 0xff000000;
        bufd[ofs] = a1;
    }

    return pipedCount;
}

extern DWORD dl3quantfast(unsigned char *inbuf, unsigned long pixels, unsigned char* userpal, DWORD colorDepth);
extern DWORD dl3quantslow(unsigned char *inbuf, unsigned long pixels, unsigned char* userpal, DWORD colorDepth);
extern void neuquantcq(unsigned char *thepic, unsigned long pixels, unsigned char* pal, unsigned long colorDepth, DWORD cycles);
static inline void strnncpy(char* dest,const char* source,DWORD maxlen)
{
    for (DWORD i = 0; i < maxlen && (dest[i] = source[i]); i++);
    dest[maxlen - 1] = 0;
}

static void setScroll(HWND hdlg,DWORD n)
{
    SCROLLINFO si;
    si.cbSize = sizeof(SCROLLINFO);
    si.fMask = SIF_POS;
    si.nPos = n;
    SetScrollInfo(GetDlgItem(hdlg, IDC_SCROLLBAR1), SB_CTL, &si,TRUE);
    WCHAR ts[4];
    ts[1] = ts[2] = ts[3] = 0;
    if (n == 100) { ts[0] = L'1'; ts[1] = L'0'; ts[2] = L'0'; }

    else {
        wchar_t t = (wchar_t)(n / 10);
        if (t) { ts[0] = t + L'0'; ts[1]=(wchar_t)n + L'0' - t*10; }
        else ts[0] = (wchar_t)n + L'0';
    }
    SetDlgItemTextW(hdlg, IDC_STATICQ, ts);
}
INT_PTR CALLBACK DlgProcgsp(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
    case WM_INITDIALOG:
    {
        LPCTSTR dlgstr[12] = { L"BLP1 JPEG",L"BLP1 Paletted", L"BLP2 JPEG",L"BLP2 Paletted 8-bit alpha",L"BLP2 DirX Compress",L"BLP2 Paletted 4-bit alpha",L"BLP2 Paletted 1-bit alpha",
            L"NeuQuant 100",L"NeuQuant 128",L"Dennis Lee3 fast",L"Dennis Lee3 slow",L"Mediancut" };
        HWND hwndl = GetDlgItem(hdlg, IDC_COMBOBOXEX1);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[0]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[1]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[2]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[3]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[4]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[5]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[6]);
        SendMessageW(hwndl, CB_SETCURSEL, 0, 0);
        hwndl = GetDlgItem(hdlg, IDC_COMBOBOXEX2);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[7]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[8]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[9]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[10]);
        SendMessageW(hwndl, CB_ADDSTRING, 0, (LPARAM)dlgstr[11]);
        SendMessageW(hwndl, CB_SETCURSEL, 0, 0);
        SetDlgItemTextW(hdlg, IDC_STATICQ, L"100");
        SetDlgItemTextW(hdlg, IDC_EDIT2, L"16");
        hwndl = GetDlgItem(hdlg, IDC_SPIN1);
        SendMessageW(hwndl, UDM_SETRANGE, 0, MAKELPARAM(16, 1));
        SendMessageW(hwndl, UDM_SETPOS, 0, MAKELPARAM(16, 0));
        hwndl = GetDlgItem(hdlg, IDC_SCROLLBAR1);
        SCROLLINFO si;
        si.cbSize = sizeof(SCROLLINFO);
        si.fMask = SIF_ALL;
        si.nMin = 1;
        si.nMax = 100;
        si.nPage = 1;
        si.nPos = 100;
        SetScrollInfo(hwndl, SB_CTL, &si, TRUE);
        return TRUE;
    }
    break;

    case WM_HSCROLL:
    {
        if (LOWORD(wParam) == SB_THUMBTRACK) {
            DWORD a = HIWORD(wParam);
            if (a < 1)a = 1;
            else if (a>100)a = 100;
            setScroll(hdlg, a);
        }
        else if (LOWORD(wParam) == SB_LINEUP) {
            SCROLLINFO si;
            si.cbSize = sizeof(SCROLLINFO);
            si.fMask = SIF_POS;
            GetScrollInfo(GetDlgItem(hdlg, IDC_SCROLLBAR1), SB_CTL, &si);
            if (si.nPos > 1)si.nPos--;
            setScroll(hdlg, si.nPos);
        }
        else if (LOWORD(wParam) == SB_LINEDOWN) {
            SCROLLINFO si;
            si.cbSize = sizeof(SCROLLINFO);
            si.fMask = SIF_POS;
            GetScrollInfo(GetDlgItem(hdlg, IDC_SCROLLBAR1), SB_CTL, &si);
            if (si.nPos < 100)si.nPos++;
            setScroll(hdlg, si.nPos);
        }
        return TRUE;
    }
    break;
    case WM_COMMAND:
        if (LOWORD(wParam) == IDOK) {
            INT_PTR a, a1;
            SCROLLINFO si;
            si.cbSize = sizeof(SCROLLINFO);
            si.fMask = SIF_POS;
            GetScrollInfo(GetDlgItem(hdlg, IDC_SCROLLBAR1), SB_CTL, &si);
            a = si.nPos;
            HWND hwndl = GetDlgItem(hdlg, IDC_COMBOBOXEX1);
            a1 = SendMessageW(hwndl, CB_GETCURSEL, 0, 0);
            a |= a1 << 8;
            if (a1 == 0 || a1 == 2) {
                if (IsDlgButtonChecked(hdlg, IDC_CHECK1))a |= 0x1000;
            }
            hwndl = GetDlgItem(hdlg, IDC_COMBOBOXEX2);
            a1 = SendMessageW(hwndl, CB_GETCURSEL, 0, 0);
            a |= a1 << 16;
            a1 = SendMessageW(GetDlgItem(hdlg, IDC_SPIN1), UDM_GETPOS, 0, 0);
            a |= a1 << 24;
            EndDialog(hdlg, a);
            return TRUE;
        }
        else if (LOWORD(wParam) == IDC_EDIT2&&HIWORD(wParam) == EN_CHANGE) {
            BOOL aa;
            UINT a = GetDlgItemInt(hdlg, IDC_EDIT2, &aa, false);
            if (a<1)SetDlgItemInt(hdlg, IDC_EDIT2, 1, false);
            if (a>16)SetDlgItemInt(hdlg, IDC_EDIT2, 16, false);
        }
        break;
    }
    return FALSE;
}


BOOL API gfpGetPluginInfo(DWORD version, LPSTR label, INT label_max_size, LPSTR extension, INT extension_max_size, INT * support)
{
    if (version != 0x0002)
        return FALSE;

    strnncpy(label, FILETYPE_BLP, label_max_size);
    strnncpy(extension, EXTENSION_BLP, extension_max_size);

    *support = GFP_READ | GFP_WRITE;

    return TRUE;
}



void * API gfpLoadPictureInit(LPCSTR filename)
{
    MYDATA * data;

    data = (MYDATA *)calloc(1, sizeof(MYDATA));
    data->header = new BLPHeader;
    data->fp = AFilemic::AOpen(filename, AFilemic::rmR);
    if (data->fp!=INVALID_HANDLE_VALUE) {
        AFilemic::ARead((BYTE*)data->header, sizeof(BLPHeader), 0, data->fp);
        if ((data->header->Signature == BLP0 || data->header->Signature == BLP1 || data->header->Signature == BLP2) &&
            data->header->Height<32769 && data->header->Height&&data->header->Width&&data->header->Width<32769 && data->header->PixelCompression<2)
            if(GetFileSize(data->fp,NULL)>0xa0)return data;
        AFilemic::AClose(data->fp);
    }
    if(data->header)delete(data->header);
    if(data)free(data);
    data = NULL;
    return data;
}

BOOL API gfpLoadPictureGetInfo(MYDATA * data, INT * pictype, INT * width, INT * height, INT * dpi, INT * bits_per_pixel, INT * bytes_per_line, BOOL * has_colormap, LPSTR label, INT label_max_size)
{
    data->width=*width = data->header->Width;
    *height= data->header->Height;
    *dpi = 72;
    *has_colormap = FALSE;
    *pictype = GFP_RGB;
    DWORD firstLenght, ImgHeaderof;

    if (data->header->Signature == BLP2) {
        data->firstMipmap = data->header->HeaderEnd.blp2.offsets[0];
        firstLenght = data->header->HeaderEnd.blp2.lengths[0];
        ImgHeaderof = 0x94U;
    }
    else {
        data->firstMipmap = data->header->HeaderEnd.blp1.offsets[0];
        firstLenght = data->header->HeaderEnd.blp1.lengths[0];
        ImgHeaderof = 0x9cU;
    }
    if (firstLenght) {
        *bits_per_pixel = 32;
      
        if (data->header->PixelCompression) {
            if (ImgHeaderof == 0x9cU) { if (!data->header->Flags || firstLenght<data->header->Width*data->header->Height)*bits_per_pixel = 8; }
            else if ((data->header->Flags & 0xff) == PALETTED && (!(data->header->Flags & 0xff00) || firstLenght<data->header->Width*data->header->Height))*bits_per_pixel = 8;
        }
        if (data->header->PixelCompression) {
            if (ImgHeaderof == 0x9CU || (data->header->Flags & 0xff) == PALETTED) {
                strnncpy(label, "BLP paletted", label_max_size);
                *pictype = GFP_BGR;
                if (data->firstMipmap>ImgHeaderof) {
                    //Paletted
                    data->filePalette = new BYTE[1024];
                    AFilemic::ARead(data->filePalette, 1024, ImgHeaderof, data->fp);
                    for (DWORD i = 0; i < 256; i++) {
                        BYTE b = data->filePalette[i * 4];
                        data->filePalette[i * 4] = data->filePalette[i * 4 + 2];
                        data->filePalette[i * 4 + 2] = b;
                    }
                    if (*bits_per_pixel == 8) {
                        data->type = 1;
                        *has_colormap = TRUE;
                    }

                    //Paletted with alpha
                    else {
                        data->type = 3;
                        *bits_per_pixel = 32;
                        data->buf = new BYTE[data->width];
                    }
                }
            }
            //dxt
            else if ((data->header->Flags & 0xff) == DXTCOMPRESSED) {
                if ((data->header->Flags & 0xff0000) == DXT3)strnncpy(label, "BLP DXT3", label_max_size);
                else if ((data->header->Flags & 0xff0000) == DXT5)strnncpy(label, "BLP DXT5", label_max_size);
                else strnncpy(label, "BLP DXT1", label_max_size);
                data->type = 4;
                data->buf = new BYTE[data->width*16];
                data->curLine = 0x80000000;
                *pictype = GFP_BGR;
                if ((data->header->Flags & 0xff0000) == DXT1)*bits_per_pixel = 24;
            }
            //rgba
            else if ((data->header->Flags & 0xff) == UNCOMPRESSED) {
                strnncpy(label, "BLP uncompressed", label_max_size);
                data->type = 5;
            }

            else return FALSE;
        }
        //JPEG
        else {
            strnncpy(label, "BLP JPEG", label_max_size);
            data->type = 6;
            data->buf = new BYTE[data->width*data->header->Height*4];

            DWORD length;
            AFilemic::ARead((BYTE*)(&length), 4, ImgHeaderof, data->fp);
            blpjpeg::jpegDecompress* pjpegDecompress = (blpjpeg::jpegDecompress*)malloc(sizeof(blpjpeg::jpegDecompress) + ((data->header->Height + 7) >> 3)*((data->width + 7) >> 3) * 512);
            if (!pjpegDecompress)return FALSE;
            BYTE* ts=pjpegDecompress->source = new BYTE[length+ firstLenght];
            if (length)AFilemic::ARead(pjpegDecompress->source, length, ImgHeaderof+4, data->fp);
            AFilemic::ARead(pjpegDecompress->source +length, firstLenght, data->firstMipmap, data->fp);
            length += firstLenght;
            pjpegDecompress->bitmap = data->buf;
            pjpegDecompress->length = length;
            pjpegDecompress->height = data->header->Height;
            pjpegDecompress->width = data->width;
            pjpegDecompress->dehuff = (short*)(pjpegDecompress + 1);
            pjpegDecompress->callback = NULL;
            blpjpeg::blpjpegDecompress(pjpegDecompress);
            delete[](ts);
            free(pjpegDecompress);

            if (!data->header->Flags) {
                *bits_per_pixel = 24;
                BYTE *source = data->buf;
                BYTE *dest = data->buf;
                for (DWORD i = data->header->Height*data->width; i--;) {
                    dest[0] = source[2];
                    dest[1] = source[1];
                    dest[2] = source[0];
                    dest += 3; source += 4;
                }
            }
            else {
                DWORD* sourced=(DWORD*)data->buf;
                for (DWORD i = data->header->Height*data->width; i--;) {
                    DWORD a = *sourced;
                    *sourced++ = (((a << 16) | (a >> 16)) & 0xff00ff) | (a & 0xff00ff00);
                }
            }
        }
        *pictype = GFP_RGB;
        data->component_per_pixel = *bits_per_pixel >>3;
        *bytes_per_line = data->component_per_pixel * *width;

        return TRUE;
    }
    return FALSE;
}

BOOL API gfpLoadPictureGetLine(MYDATA * data, INT line, unsigned char * buffer)
{
    if (data->type ==1)AFilemic::ARead(buffer, data->width, data->firstMipmap + line*data->width, data->fp);
    else if (data->type <=3) {
        AFilemic::ARead(data->buf, data->width, data->firstMipmap + line*data->width, data->fp);
        DWORD* destd = (DWORD*)buffer;
        for (DWORD i = 0; i<(DWORD)data->width; i++)*destd++ = *(DWORD*)(data->filePalette + data->buf[i] * 4);
        if (data->type == 3) {
            DWORD alphaDepth = (data->header->Flags & 0xf00) >> 8;
            DWORD a = data->width*line;
            BYTE bof = (BYTE)(a & 7);
            if (alphaDepth == 4)a >>= 1;
            else if (alphaDepth == 2)a >>= 2;
            else if (alphaDepth == 1)a >>= 3;
            AFilemic::ARead(data->buf, data->width, a + data->firstMipmap + data->header->Height*data->width, data->fp);
            BYTE* source = data->buf;
            if (alphaDepth == 8) {
                for (DWORD i = 0; i<(DWORD)data->width; i++)buffer[i * 4 + 3] = source[i];
            }
            else if (alphaDepth == 4) {
                bof &= 1;
                for (DWORD i = 0; i < (DWORD)data->width; i++) {
                    BYTE b = *source;
                    if (bof) { b = b & 0xf0; b |= b >> 4; source++; }
                    else { b = b & 0xf; b |= b << 4; }
                    bof = !bof;
                    buffer[i * 4 + 3] = b;
                }
            }
            else if (alphaDepth == 2) {
                bof = 1 << (bof & 3);
                BYTE b = *source;
                for (DWORD i = 0; i<(DWORD)data->width; i++) {
                    BYTE b1 = 0;
                    if (b&bof)b = 0x55;
                    bof += bof;
                    if (a&bof)b |= 0xaa;
                    bof += bof;
                    if (!bof) { bof = 1; source++; b = *source; }
                    buffer[i * 4 + 3] = b1;
                }
            }
            else if (alphaDepth == 1) {
                bof = 1 << bof;
                BYTE b = *source;
                for (DWORD i = 0; i<(DWORD)data->width; i++) {
                    if (b&bof)buffer[i * 4 + 3] = 0xff;
                    else buffer[i * 4 + 3] = 0;
                    bof += bof;
                    if (!bof) { bof = 1; source++; b = *source; }
                }
            }
        }
    }
    else if (data->type == 5) {
        AFilemic::ARead(buffer, data->width, data->firstMipmap + line*data->width * 4, data->fp);
        for (DWORD i = data->width; i--;) {
            BYTE b = *buffer;
            *buffer = *(buffer+2);
            *(buffer + 2) = b;
            buffer += 4;
        }
    }
    else if (data->type == 6) { 
        line = data->header->Height - line - 1;
        if (!data->header->Flags) memcpy(buffer, data->buf + line*data->width * 3, data->width * 3);
        else memcpy(buffer, data->buf + line*data->width * 4, data->width * 4); 
    }
    else if (data->type == 4) {
        DWORD a = (DWORD)line - data->curLine;
        
        if (a < 4UL) { 
            if ((data->header->Flags & 0xff0000) != DXT1)memcpy(buffer, data->buf + a*data->width * 4, data->width * 4);
            else memcpy(buffer, data->buf + a*data->width * 4, data->width * 3); 
        }
        else {
            data->curLine = (DWORD)line & 0xfffffffc;
            a=(data->width+3) & 0xfffffffc;
            if ((data->header->Flags & 0xff0000) == DXT1)a >>= 1;
            if(data->fp != INVALID_HANDLE_VALUE)SetFilePointer(data->fp, data->firstMipmap + a*data->curLine, NULL, FILE_BEGIN);
            BYTE unp[64];
            BYTE source[16];
            a = 1;
            unsigned long readCount;
            if ((data->header->Flags & 0xff0000) == DXT3) a = 3;
            else if ((data->header->Flags & 0xff0000) == DXT5) a = 5;
            if (a == 1) {
                for (DWORD x = 0; x < (DWORD)data->width; x++) {
                    DWORD t = (x & 3)<<2;
                    if (!t) {
                        if (data->fp != INVALID_HANDLE_VALUE)ReadFile(data->fp, source, 8, &readCount, NULL);
                        dxt::DecompressDxt(unp, source, a);
                    }
                    DWORD *destd = (DWORD*)(data->buf + x * 3);
                    *destd = *(DWORD*)(unp+t);
                    destd += data->width;
                    *destd = *(DWORD*)(unp + t+16);
                    destd += data->width;
                    *destd = *(DWORD*)(unp + t+32);
                    destd += data->width;
                    *destd = *(DWORD*)(unp + t+48);
                }
                a = (DWORD)line - data->curLine;
                memcpy(buffer, data->buf + a*data->width * 4, data->width * 3);
            }
            else {
                for (DWORD x = 0; x < (DWORD)data->width; x++) {
                    DWORD t = (x & 3)<<2;
                    if (!t) {
                        if (data->fp != INVALID_HANDLE_VALUE)ReadFile(data->fp, source, 16, &readCount, NULL);
                        dxt::DecompressDxt(unp, source, a);
                    }
                    DWORD *destd = (DWORD*)(data->buf + x * 4);
                    *destd= *(DWORD*)(unp + t);
                    destd += data->width;
                    *destd = *(DWORD*)(unp + t+16);
                    destd += data->width;
                    *destd = *(DWORD*)(unp + t+32);
                    destd += data->width;
                    *destd = *(DWORD*)(unp + t+48);
                }
                a = (DWORD)line - data->curLine;
                memcpy(buffer, data->buf + a*data->width * 4, data->width * 4);
            }
           
        }
    }
    return TRUE;
}

BOOL API gfpLoadPictureGetColormap(MYDATA * data, GFP_COLORMAP * cmap)
{
    if (data->filePalette) {
        for (DWORD i = 0; i<256;i++) {
            cmap->red[i] = data->filePalette[i*4];
            cmap->green[i] = data->filePalette[i * 4+1];
            cmap->blue[i] = data->filePalette[i * 4+2];
        }
        return TRUE;
    }
    else return FALSE;
}

void API gfpLoadPictureExit(MYDATA * data)
{
    AFilemic::AClose(data->fp);
    if (data->filePalette)delete[](data->filePalette);
    if (data->buf)delete[](data->buf);
    if (data->header)delete(data->header);
    free(data);
}


BOOL API gfpSavePictureIsSupported(INT width, INT height, INT bits_per_pixel, BOOL has_colormap)
{
    return(bits_per_pixel==32|| bits_per_pixel==24||bits_per_pixel==8|| (bits_per_pixel < 5&& has_colormap)|| bits_per_pixel == 1);
}

void API gfpSavePicturePutColormap(MYDATA * data, GFP_COLORMAP * cmap)
{
    if (cmap) {
        DWORD c;
        BYTE* source = (BYTE*)cmap;
        if (data->bits_per_pixel >4)c = 256;
        else c = 1<< data->bits_per_pixel;
        for (DWORD i=0; i<c; i++) {
            data->filePalette[i * 4] = source[c+c];
            data->filePalette[i * 4 + 1] = source[c];
            data->filePalette[i * 4 + 2] = *source;
            data->filePalette[i * 4 + 3] = 255;
            data->type = 1;
            source++;
        }
    }
    //Transparensy index unknown
}

MYDATA * API gfpSavePictureInit(LPCSTR filename, INT width, INT height, INT bits_per_pixel, INT dpi, INT * picture_type, LPSTR label, INT label_max_size)
{
    MYDATA * data;
    data = (MYDATA *)calloc(1, sizeof(MYDATA));
    data->fp=AFilemic::AOpen(filename, AFilemic::rmW);
    if (data->fp == INVALID_HANDLE_VALUE) {
        free(data);
        data = NULL;
    }
    else {
        *picture_type = GFP_RGB;
        data->filePalette = (BYTE*)malloc(width*height * 4 + 0x400);
        data->buf = data->filePalette + 0x400;
        data->header = (BLPHeader*)malloc(width*height * 8 + 0x1800);
        data->type = 0;
        strnncpy(label, FILETYPE_BLP, label_max_size);
        data->header->Height = height;
        data->header->Width = data->width = width;
        data->bits_per_pixel = bits_per_pixel;
        data->component_per_pixel = (bits_per_pixel+7) / 8; 
    }
    return data;
}

BOOL API gfpSavePicturePutLine(MYDATA * data, INT line, const unsigned char * buffer)
{
    BYTE* dest = data->buf + line*data->width * 4;
    DWORD *destd = (DWORD*)dest;
    switch (data->bits_per_pixel) {
    case 1: {
        DWORD c1, c2;
        if (data->type) {
            c1 = *(DWORD*)data->filePalette; c2 = *(DWORD*)(data->filePalette + 4);
        }
        else {
            c1 = 0xff000000UL; c2 = 0xffffffff;
        }
        BYTE b, bof = 0;
        for (DWORD x = 0; x < (DWORD)data->width; x++) {
            if (!bof) {
                bof = 0x80; b = *buffer++;
            }
            if (b&bof)destd[x] = c2;
            else destd[x] = c1;
            bof >>= 1;
        }
    }
            break;
    case 2:
    {
        BYTE bof = 0;
        BYTE b;
        DWORD* source = (DWORD*)data->filePalette;
        for (DWORD x = 0; x < (DWORD)data->width; x++) {
            if (!bof) {
                bof=4; b = *buffer++;
            }
            destd[x] = source[b >> 6];
            b<<=2;
            bof--;
        }
    }
        break;
    case 4:
    case 3:
     {
        BYTE bof = 1;
        BYTE b;
        DWORD* source = (DWORD*)data->filePalette;
        for (DWORD x = 0; x < (DWORD)data->width; x++) {
            if (bof) b = *buffer++;
            if (bof) destd[x]= source[b >>4];
            else destd[x] = source[b & 0xf];
            bof = !bof;
        }
    }
        break;
    case 8: {
        if(data->type)
            for (DWORD x = 0; x < (DWORD)data->width; x++)destd[x] = *(DWORD*)(data->filePalette+4*(buffer[x]));
        else
            for (DWORD x = 0; x < (DWORD)data->width; x++)destd[x] =  0x10101UL * buffer[x]+0xff000000UL;
    }
            break;
    case 24: {
        DWORD a;
        for (DWORD x = data->width - 1; x--;) {
            a = *(DWORD*)buffer;
            a |= 0xff000000;
            buffer += 3;
            *destd++ = a;
        }
        a = *(DWORD*)(buffer - 1);
        a = (a >> 8) | 0xff000000;
        *destd = a;
    }
             break;
    case 32: {memcpy(dest, buffer, data->width * 4);}
             break;

    }
    
    return TRUE;
}

void API gfpSavePictureExit(MYDATA * data)
{
    DWORD saveparam;
    saveparam = (DWORD)DialogBoxParamW(hDLLModule, MAKEINTRESOURCE(IDD_DIALOG1), GetActiveWindow(), &DlgProcgsp, 0L);
    if (!saveparam)saveparam = 0x10000064;
    BYTE savef = (BYTE)((saveparam & 0xf00)>>8);
    if (savef < 2)data->header->Signature = BLP1;
    else data->header->Signature = BLP2;
    if (savef == 0 || savef == 2)data->header->PixelCompression = 0;
    else data->header->PixelCompression = 1;
    DWORD alphaDepth = 0;
    BYTE* source,*dest;
    if (data->component_per_pixel == 4) {
        alphaDepth = 8;
        if(savef==5)alphaDepth = 4;
        else if (savef == 6)alphaDepth = 1;
    }


    dest = (BYTE*)data->header;
    data->header->Flags = 0;
    if (savef < 2) {
        dest += 0x9c;
        data->header->HeaderEnd.blp1.flags2 = 1;
        if (alphaDepth) {
            data->header->Flags = 8;
            data->header->HeaderEnd.blp1.alphaEncoding = 4;
        }
        else data->header->HeaderEnd.blp1.alphaEncoding = 5;
    }
    else {
        dest += 0x94;
        DWORD v = 0x1000000 + (alphaDepth << 8);
        if (savef == 3|| savef == 5|| savef == 6)v |= 0x80001;//blp2 paletted
        else if (savef == 4) {//dxt
            if (alphaDepth)v |= 0x70002;
            else v |= 0x2;
        }
        data->header->Flags = v;
    }
    if ((savef == 4)==(data->component_per_pixel==1)) {
        source = data->buf;
        for (DWORD i = data->width*data->header->Height; i--;) {
            BYTE b = *source;
            *source = *(source + 2);
            *(source + 2) = b;
            source += 4;
        }
    }
    if (savef == 1 || savef == 3 || savef == 5 || savef == 6) {
        if (data->component_per_pixel>1) {
            DWORD qw = saveparam&0xf0000,l=data->header->Height*data->width;
            if (qw == 0x20000) {
                qw = dl3quantfast(data->buf, l, data->filePalette, 4);
                if (qw<256)GetPaletteMCCQ(data->filePalette, (piped*)dest, data->buf, l, 4);
            }
            else if (qw == 0x30000) {
                qw = dl3quantslow(data->buf, l, data->filePalette, 4);
                if (qw<256)GetPaletteMCCQ(data->filePalette, (piped*)dest, data->buf, l, 4);
            }
            else if (!qw && l>503)neuquantcq(data->buf, l, data->filePalette, 4, 100);
            else if (qw == 0x10000 && l>503)neuquantcq(data->buf, l, data->filePalette, 4, 128);
            else GetPaletteMCCQ(data->filePalette, (piped*)dest, data->buf, l, 4);
        }
        memcpy(dest, data->filePalette, 0x400);
        dest += 0x400;
    }

    DWORD* offs,*lens;
    if (savef < 2) {
        offs = data->header->HeaderEnd.blp1.offsets;
        lens = data->header->HeaderEnd.blp1.lengths;
    }
    else {
        offs = data->header->HeaderEnd.blp2.offsets;
        lens = data->header->HeaderEnd.blp2.lengths;
    }

    blpjpeg::jpegCompress* pjpegCompress;
    if (savef == 0 || savef == 2) {
        //JPEG
        if (savef == 0)offs[0] = 0xa0;
        else offs[0] = 0x98;
        *(DWORD*)dest = 0;
        dest += 4;
        pjpegCompress = (blpjpeg::jpegCompress*)malloc(sizeof(blpjpeg::jpegCompress) + ((data->header->Height + 7) >> 3)*((data->width + 7) >> 3) * 512);
        pjpegCompress->quality = saveparam&0xff;
    }
    else {
        if (savef == 1)offs[0] = 0x49c;
        else if (savef == 3 || savef == 5 || savef == 6)offs[0] = 0x494;
        else offs[0] = 0x94;
    }

    //convert
    DWORD height = data->header->Height;
    DWORD width = data->width;
    DWORD length,a;
    DWORD currentI;
    DWORD maxI = saveparam >> 24;
    for (currentI = 0; ;) {
        DWORD* sourced, *destd,s;

        //JPEG
        if (savef == 0 || savef == 2) {
            
            pjpegCompress->height = height;
            pjpegCompress->width = width;
            pjpegCompress->inbuff = data->buf;
            pjpegCompress->workbuff = (short*)(pjpegCompress + 1);
            pjpegCompress->outbuff = dest;
            if ((height*width)<4096 || !(saveparam & 0xf000))
                blpjpeg::blpjpegCompressBase(pjpegCompress);
            else
                blpjpeg::blpjpegCompressProgressive(pjpegCompress);
            //length
            length = pjpegCompress->outlenght;
            dest += length;
        }
        //paletted
        else if ((savef == 1 || savef == 3 || savef == 5 || savef == 6)) {

            sourced = (DWORD*)data->buf;
            length = height*width;
            BYTE* desta = dest + length;
            BYTE bof = 1;
            DWORD min, minj;
            long var;
            BYTE b;

            if (alphaDepth == 1)*desta = 0;
            for (DWORD i = length; i--; sourced++, dest++) {
                a = *sourced;

                source = data->filePalette;
                min = 0xffff;
                for (DWORD j = 0; j<256; j++) {
                    b = (BYTE)a;
                    var = b - *source++;
                    s = var * var;
                    b = (BYTE)(a >> 8);
                    var = b - *source++;
                    s += var * var;
                    b = (BYTE)(a >> 16);
                    var = b - *source++;
                    s += var * var;
                    source++;
                    if (s<min) { min = s; minj = j; }
                }

                *dest = (BYTE)minj;
                //alpha
                if (alphaDepth) {
                    if (alphaDepth == 8)*desta++ = (BYTE)(a >> 24);
                    else if (alphaDepth == 4) {
                        if (bof) {
                            BYTE b = (BYTE)(a >> 28);
                            b &= 0xf;
                            *desta = b;
                        }
                        else {
                            BYTE b = (BYTE)(a >> 24);
                            b &= 0xf0;
                            *desta |= b;
                            desta++;
                        }
                        bof = !bof;
                    }
                    else if (alphaDepth == 1) {
                        if ((long)a<0)*desta |= bof;
                        bof += bof;
                        if (!bof) {
                            bof++; desta++; *desta = 0;
                        }
                    }
                }
            }
            sourced = (DWORD*)data->buf;
            
            if (alphaDepth == 1)a= (length + 7) >> 3;
            else if (alphaDepth == 4)a= (length + 1) >> 1;
            else if (alphaDepth == 8)a= length;
            else a = 0;
            dest += a;
            //length
            length += a;
        }
        //dxt
        else {

            //buf+400=>dest
            sourced = (DWORD*)data->buf;
            destd = (DWORD*)data->filePalette;
            for (DWORD y = 0; y<height; y += 4) {
                for (DWORD x = 0; x<width; x += 4) {
                    *destd++ = *(sourced + x + y*width);
                    *destd++ = *(sourced + x + 1 + y*width);
                    *destd++ = *(sourced + x + 2 + y*width);
                    *destd++ = *(sourced + x + 3 + y*width);
                    *destd++ = *(sourced + x + (y + 1)*width);
                    *destd++ = *(sourced + x + 1 + (y + 1)*width);
                    *destd++ = *(sourced + x + 2 + (y + 1)*width);
                    *destd++ = *(sourced + x + 3 + (y + 1)*width);
                    *destd++ = *(sourced + x + (y + 2)*width);
                    *destd++ = *(sourced + x + 1 + (y + 2)*width);
                    *destd++ = *(sourced + x + 2 + (y + 2)*width);
                    *destd++ = *(sourced + x + 3 + (y + 2)*width);
                    *destd++ = *(sourced + x + (y + 3)*width);
                    *destd++ = *(sourced + x + 1 + (y + 3)*width);
                    *destd++ = *(sourced + x + 2 + (y + 3)*width);
                    *destd++ = *(sourced + x + 3 + (y + 3)*width);
                    destd -= 16;
                    int mask;
                    a = width - x;
                    if (a == 1)mask = 0x1111;
                    else if (a == 2)mask = 0x3333;
                    else if (a == 3)mask = 0x7777;
                    else mask = 0xffff;
                    a = height - y;
                    if (a == 1)mask &= 0xf;
                    else if (a == 2)mask &= 0xff;
                    else if (a == 3)mask &= 0xfff;
                    if (alphaDepth) {
                        squish::CompressMasked((BYTE*)destd, mask, dest, squish::kDxt5);
                        dxt::CompressAlphaDxt5((BYTE*)destd, mask, dest);
                        dest += 16;
                    }
                    else {
                        squish::CompressMasked((BYTE*)destd, mask, dest, squish::kDxt1);
                        dest += 8;
                    }
                }
            }
            //length
            if (alphaDepth)length = ((height + 3) & 0xfffffffc)*((width + 3) & 0xfffffffc);
            else length = ((height + 3) >> 2)*((width + 3) >> 2) * 8;
        }

        lens[currentI] = length;
        if ((width + height)>2) {
            length += offs[currentI];
            offs[++currentI] = length;
            if (currentI == maxI) {
                currentI--;
                break;
            }
        }
        else break;
        //resize
        sourced = destd = (DWORD*)data->buf;
        if (width == 1) {
            if (height != 1) {
                height = height >> 1;
                for (DWORD y = 0; y<height; y++) {
                    s = *sourced++;
                    a = (s & 0xff00ff00) >> 8;
                    DWORD a1 = s & 0xff00ff;
                    s = *sourced++;
                    a += (s & 0xff00ff00) >> 8;
                    a1 += s & 0xff00ff;
                    destd[y] = ((a << 7) & 0xff00ff00) + ((a1 >> 1) & 0xff00ff);
                }
            }
        }
        else if (height == 1) {
            if (width != 1) {
                width = width >> 1;
                for (DWORD x = 0; x<width; x++) {
                    s = *sourced++;
                    a = (s & 0xff00ff00) >> 8;
                    DWORD a1 = s & 0xff00ff;
                    s = *sourced++;
                    a += (s & 0xff00ff00) >> 8;
                    a1 += s & 0xff00ff;
                    destd[x] = ((a << 7) & 0xff00ff00) + ((a1 >> 1) & 0xff00ff);
                }
            }
        }
        else {
            height = height >> 1;
            for (DWORD y = 0; y<height; y++) {
                for (DWORD x = 0; x<(width >> 1); x++) {
                    s = *sourced;
                    a = (s & 0xff00ff00) >> 8;
                    DWORD a1 = s & 0xff00ff;
                    s = *(sourced + width);
                    a += (s & 0xff00ff00) >> 8;
                    a1 += s & 0xff00ff;
                    sourced++;
                    s = *sourced;
                    a += (s & 0xff00ff00) >> 8;
                    a1 += s & 0xff00ff;
                    s = *(sourced + width);
                    a += ((s & 0xff00ff00) >> 8) + 1;
                    a1 += (s & 0xff00ff) + 1;
                    *destd++ = ((a << 6) & 0xff00ff00) + ((a1 >> 2) & 0xff00ff);
                    sourced++;
                }
                if (width & 1)sourced++;
                sourced += width;
            }
            width = width >> 1;
        }
    }
    //compress the jpg headers
    if ((savef == 0 || savef == 2) && currentI>1) {
        dest = (BYTE*)data->header;
        DWORD bn;
        for (bn = 0; bn < lens[currentI]; bn++) {
            BYTE cb = *(dest + offs[0]+bn);
            DWORD i;
            for (i = currentI; i; i--) {
                if (cb != *(dest + offs[i] + bn))break;
            }
            if (i)break;
        }
        if(savef == 0)*(DWORD*)(dest + 0x9c) = bn;
        else *(DWORD*)(dest + 0x94) = bn;
        lens[0] -= bn;
        offs[0] += bn;
        for (DWORD i = 1; i <= currentI; i++) {
            lens[i] -= bn;
            memmove(dest + offs[i] - bn*(i-1), dest + offs[i]+bn, lens[i]);
            offs[i] -= bn*(i-1);
        }
    }

    a = offs[currentI] + lens[currentI];

    for (; (height > 1) || (width > 1);) {
        height >>= 1; width >>= 1;
        ++currentI;
        offs[currentI] = offs[currentI - 1];
        lens[currentI] = lens[currentI - 1];
    }
    for (; ++currentI < 16;)offs[currentI] = lens[currentI] = 0;
    AFilemic::AWrite((BYTE*)data->header, a, data->fp);
    
    if (data->filePalette)free(data->filePalette);
    if (data->header)free(data->header);
    AFilemic::AClose(data->fp);
    free(data);

}

