#pragma once

namespace dxt {
    typedef unsigned long long int QWORD;
    typedef unsigned char BYTE;
    typedef unsigned long DWORD;
    typedef unsigned short WORD;
    void DecompressAlphaDxt3sub(BYTE* rgba, DWORD half)
    {
        DWORD l = half & 0xf0f0f0f;
        half = half^l;
        half = half | (half >> 4);
        l = l | (l << 4);
        rgba[3] = (BYTE)l;
        rgba[7] = (BYTE)half;
        half = half >> 8;
        l = l >> 8;
        rgba[11] = (BYTE)l;
        rgba[15] = (BYTE)half;
        half = half >> 8;
        l = l >> 8;
        rgba[19] = (BYTE)l;
        rgba[23] = (BYTE)half;
        half = half >> 8;
        l = l >> 8;
        rgba[27] = (BYTE)l;
        rgba[31] = (BYTE)half;
    }
    void DecompressAlphaDxt3(BYTE* rgba, QWORD h)
    {
        QWORD l = h & 0xf0f0f0f0f0f0f0f;
        h = h^l;
        h = h | (h >> 4);
        l = l | (l << 4);
        rgba[3] = (BYTE)l;
        rgba[7] = (BYTE)h;
        for (int i = 7; i--;) {
            rgba += 8;
            h = h >> 8;
            l = l >> 8;
            rgba[3] = (BYTE)l;
            rgba[7] = (BYTE)h;
        }
    }

    void DecompressAlphaDxt532(BYTE* rgba, const BYTE* block)
    {
        // get the two alpha values

        DWORD alpha0 = block[0];
        DWORD alpha1 = block[1];

        // compare the values to build the codebook
        BYTE codes[8];
        codes[0] = (BYTE)alpha0;
        codes[1] = (BYTE)alpha1;

        if (alpha0 <= alpha1) {
            alpha1 -= alpha0;
            codes[2] = (BYTE)(alpha0 + alpha1 / 5);
            codes[3] = (BYTE)(alpha0 + (2 * alpha1) / 5);
            codes[4] = (BYTE)(alpha0 + (3 * alpha1) / 5);
            codes[5] = (BYTE)(alpha0 + (4 * alpha1) / 5);
            codes[6] = 0;
            codes[7] = 255;

        }
        else {
            alpha0 -= alpha1;
            codes[7] = (BYTE)(alpha1 + alpha0 / 7);
            codes[6] = (BYTE)(alpha1 + (2 * alpha0) / 7);
            codes[5] = (BYTE)(alpha1 + (3 * alpha0) / 7);
            codes[4] = (BYTE)(alpha1 + (4 * alpha0) / 7);
            codes[3] = (BYTE)(alpha1 + (5 * alpha0) / 7);
            codes[2] = (BYTE)(alpha1 + (6 * alpha0) / 7);
        }



        block = block + 2;
        DWORD value = *((DWORD*)block);
        block += 2;
        for (int j = 3; j < 35; j += 4) {
            rgba[j] = codes[value & 0x7];
            value = value >> 3;
        }
        value = *((int*)block);
        value = value >> 8;
        for (int j = 35; j < 67; j += 4) {
            rgba[j] = codes[value & 0x7];
            value = value >> 3;
        }
    }
    void DecompressAlphaDxt5(BYTE* rgba, QWORD block)
    {
        BYTE codes[8];
        *(WORD*)codes = (WORD)block;
        DWORD alpha0 = (BYTE)block;
        block >>= 8;
        DWORD alpha1 = (BYTE)block;
        block >>= 8;
        if (alpha0 <= alpha1) {
            alpha1 -= alpha0;
            codes[2] = (BYTE)(alpha0 + alpha1 / 5);
            codes[3] = (BYTE)(alpha0 + (2 * alpha1) / 5);
            codes[4] = (BYTE)(alpha0 + (3 * alpha1) / 5);
            codes[5] = (BYTE)(alpha0 + (4 * alpha1) / 5);
            codes[6] = 0;
            codes[7] = 255;
        }
        else {
            alpha0 -= alpha1;
            codes[7] = (BYTE)(alpha1 + alpha0 / 7);
            codes[6] = (BYTE)(alpha1 + (2 * alpha0) / 7);
            codes[5] = (BYTE)(alpha1 + (3 * alpha0) / 7);
            codes[4] = (BYTE)(alpha1 + (4 * alpha0) / 7);
            codes[3] = (BYTE)(alpha1 + (5 * alpha0) / 7);
            codes[2] = (BYTE)(alpha1 + (6 * alpha0) / 7);
        }
        for (int j = 3; j < 67; j += 4) {
            rgba[j] = codes[block & 0x7];
            block >>= 3;
        }
    }

    void DecompressColour(BYTE* rgba, BYTE const* block, bool isDxt1)
    {
        BYTE codes[16];
        DWORD* colour = (DWORD*)codes;
        DWORD a = *(DWORD*)block;
        DWORD c;
        colour[0] = colour[1] = colour[2] = colour[3] = 0xff000000;
        c = a & 0xf800f800;
        c >>= 8;
        c |= c >> 5;
        codes[0] = (BYTE)c;
        codes[4] = (BYTE)(c >> 16);
        c = a & 0x07e007e0;
        c >>= 3;
        c |= c >> 6;
        codes[1] = (BYTE)c;
        codes[5] = (BYTE)(c >> 16);
        c = a & 0x1f001f;
        c = c << 3 | c >> 2;
        codes[2] = (BYTE)c;
        codes[6] = (BYTE)(c >> 16);
        if (isDxt1 && ((WORD)a <= (WORD)(a >> 16))) {
            for (int i = 0; i<3; i++) {
                a = codes[i];
                c = codes[4 + i];
                codes[8 + i] = (BYTE)((a + c) / 2);
            }
        }
        else {
            for (int i = 0; i<3; i++) {
                a = codes[i];
                c = codes[4 + i];
                codes[8 + i] = (BYTE)((a + a + c) / 3);
                codes[12 + i] = (BYTE)((a + c + c) / 3);
            }

        }
        a = *(DWORD*)(block + 4);
        for (int i = 0; i<16; i++) {
            *(DWORD*)(rgba + 4 * i) = colour[a & 3];
            a >>= 2;
        }

    }


    void DecompressDxt(BYTE* rgba, BYTE const* block, int DxtNum)
    {
        BYTE codes[16];
        DWORD* colour = (DWORD*)codes;
        DWORD a;
        if (DxtNum - 1)a = *(DWORD*)(block + 8);
        else a = *(DWORD*)block;
        DWORD c;
        colour[0] = colour[1] = colour[2] = colour[3] = 0xff000000;
        c = a & 0xf800f800;
        c >>= 8;
        c |= c >> 5;
        codes[0] = (BYTE)c;
        codes[4] = (BYTE)(c >> 16);
        c = a & 0x07e007e0;
        c >>= 3;
        c |= c >> 6;
        codes[1] = (BYTE)c;
        codes[5] = (BYTE)(c >> 16);
        c = a & 0x1f001f;
        c = c << 3 | c >> 2;
        codes[2] = (BYTE)c;
        codes[6] = (BYTE)(c >> 16);
        if (DxtNum == 1 && ((WORD)a <= (WORD)(a >> 16))) {
            for (int i = 0; i<3; i++) {
                a = codes[i];
                c = codes[4 + i];
                codes[8 + i] = (BYTE)((a + c) / 2);
            }
        }
        else {
            for (int i = 0; i<3; i++) {
                a = codes[i];
                c = codes[4 + i];
                codes[8 + i] = (BYTE)((a + a + c) / 3);
                codes[12 + i] = (BYTE)((a + c + c) / 3);
            }

        }
        if (DxtNum - 1)a = *(DWORD*)(block + 12);
        else a = *(DWORD*)(block + 4);
        for (int i = 0; i<16; i++) {
            *(DWORD*)(rgba + 4 * i) = colour[a & 3];
            a >>= 2;
        }
        if (DxtNum == 3) {
#ifdef _M_X64
            dxt::DecompressAlphaDxt3(rgba, *(QWORD*)block);
#else
            dxt::DecompressAlphaDxt3sub(rgba, *(DWORD*)block);
            dxt::DecompressAlphaDxt3sub(rgba + 32, *(DWORD*)(block + 4));
#endif
        }
        else if (DxtNum == 5) {
#ifdef _M_X64
            dxt::DecompressAlphaDxt5(rgba, *(QWORD*)block);
#else
            dxt::DecompressAlphaDxt532(rgba, block);
#endif
        }
    }


    void CompressAlphaDxt5(BYTE const* rgba, DWORD mask, BYTE* outblock)
    {
        DWORD aval[16];
        DWORD c, alpha0, alpha1, min5, max5, a, min, max, mins;
        c = 1;
        aval[0] = min = min5 = max = max5 = rgba[3];
        a = 1;
        for (DWORD i = 7; i < 67; i += 4) {
            a <<= 1;
            if (a & mask) {
                aval[c++] = rgba[i];
                if (min > rgba[i])min = rgba[i];
                if (max < rgba[i])max = rgba[i];
                if (min5 > rgba[i] && (rgba[i]>8 || min5<9))min5 = rgba[i];
                if (max5 < rgba[i] && (rgba[i]<0xf7 || max5>0xf6))max5 = rgba[i];
            }
        }
        for (DWORD i = 0; c < 16; i++)aval[c++] = aval[i];
        mins = 0xffffffff;
        if (min > 3 || max < 0xfb) {
            //7
            a = min + 16;
            if ((a + 8) > max) {
                a = max - 8 - ((a^max) & 1);
                if (a > max)a = min;
            }

            DWORD b;
            b = max - 16;
            if ((b - 16)<min || b>max) {
                b = min + 16 + ((a^max) & 1);
                if (b > max)b = max;
            }
            for (DWORD i = a; i < 0x100; i -= 2) {
                for (a = b; a < 0x100; a += 2) {
                    DWORD cval[8];
                    for (DWORD k = 0; k < 8; k++)cval[k] = i + (k*(a - i)) / 7;
                    DWORD s = 0;
                    for (DWORD k = 0; k < 16; k++) {
                        long d = aval[k] - cval[0];
                        DWORD m;
                        if (d < 0)d = -d;
                        m = (DWORD)d;
                        d = aval[k] - cval[1];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[2];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[3];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[4];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[5];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[6];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[7];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        s += m*m;
                    }
                    if (mins > s) {
                        mins = s;
                        alpha1 = i;
                        alpha0 = a;
                    }
                }
            }
        }
        if (min < 9 || max > 0xf6) {
            //5
            if (min5 > max5) { a = min5; min5 = max5; max5 = a; }
            a = min5 + 18;
            if ((a + 6) > max5) {
                a = max5 - 6 - ((a^max5) & 1);
                if (a > max5)a = min5;
            }

            DWORD b;
            b = max5 - 18;
            if ((b - 16)<min5 || b>max5) {
                b = min5 + 16 + ((a^max5) & 1);
                if (b > max5)b = max5;
            }
            for (DWORD i = a; i < 0x100; i -= 2) {
                for (a = b; a < 0x100; a += 2) {
                    DWORD cval[8];
                    for (DWORD k = 0; k < 6; k++)cval[k] = i + (k*(a - i)) / 5;
                    cval[6] = 0;
                    cval[7] = 255;
                    DWORD s = 0;
                    for (DWORD k = 0; k < 16; k++) {
                        long d = aval[k] - cval[0];
                        DWORD m;
                        if (d < 0)d = -d;
                        m = (DWORD)d;
                        d = aval[k] - cval[1];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[2];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[3];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[4];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[5];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[6];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        d = aval[k] - cval[7];
                        if (d < 0)d = -d;
                        if ((DWORD)d < m)m = (DWORD)d;
                        s += m*m;
                    }
                    if (mins > s) {
                        mins = s;
                        alpha1 = a;
                        alpha0 = i;
                    }
                }
            }
        }


        outblock[0] = (BYTE)alpha0;
        outblock[1] = (BYTE)alpha1;
        DWORD codes[8];
        codes[0] = alpha0;
        codes[1] = alpha1;
        if (alpha0 <= alpha1) {
            alpha1 -= alpha0;
            codes[2] = (BYTE)(alpha0 + alpha1 / 5);
            codes[3] = (BYTE)(alpha0 + (2 * alpha1) / 5);
            codes[4] = (BYTE)(alpha0 + (3 * alpha1) / 5);
            codes[5] = (BYTE)(alpha0 + (4 * alpha1) / 5);
            codes[6] = 0;
            codes[7] = 255;
        }
        else {
            alpha0 -= alpha1;
            codes[7] = (BYTE)(alpha1 + alpha0 / 7);
            codes[6] = (BYTE)(alpha1 + (2 * alpha0) / 7);
            codes[5] = (BYTE)(alpha1 + (3 * alpha0) / 7);
            codes[4] = (BYTE)(alpha1 + (4 * alpha0) / 7);
            codes[3] = (BYTE)(alpha1 + (5 * alpha0) / 7);
            codes[2] = (BYTE)(alpha1 + (6 * alpha0) / 7);
        }

        for (DWORD i = 16; i--; ) {
            a <<= 3;
            if (mask&(1 << i)) {
                DWORD al = rgba[4 * i + 3];
                long d = al - codes[0];
                if (d < 0)d = -d;
                min = (DWORD)d;
                c = 0;
                for (DWORD j = 1; j < 8; j++) {
                    d = al - codes[j];
                    if (d < 0)d = -d;
                    if ((DWORD)d < min) {
                        min = (DWORD)d;
                        c = j;
                    }
                }
                a += c;
            }
            if (i == 8)*(DWORD*)(outblock + 4) = (a << 8);

        }

        *(DWORD*)(outblock + 2) = a;
    }




}
