/* NeuQuant Neural-Net Quantization Algorithm
* ------------------------------------------
*
* Copyright (c) 1994 Anthony Dekker
*
*/




typedef long pixel[4];				/* BGRc */
static pixel network[256];			/* the network itself */

static long netindex[256];			/* for network lookup - really 256 */

static long bias[256];			/* bias and freq arrays for learning */
static long freq[256];
static long radpower[32];			/* radpower for precomputation */



                                    /* Search for biased BGR values
                                    ---------------------------- */

static long contest(long r, long g, long b)
{
    /* finds closest neuron (min dist) and updates freq */
    /* finds best neuron (min dist-bias) and returns position */
    /* for frequently chosen neurons, freq[i] is high and bias[i] is negative */
    /* bias[i] = gamma*((1/netsize)-freq[i]) */

    long i, dist, a, biasdist, betafreq;
    long bestpos, bestbiaspos, bestd, bestbiasd;
    long *p, *f, *n;

    bestd = 0x7fffffff;
    bestbiasd = bestd;
    bestpos = -1;
    bestbiaspos = bestpos;
    p = bias;
    f = freq;

    for (i = 0; i<256; i++) {
        n = network[i];
        dist = n[0] - r;   if (dist<0) dist = -dist;
        a = n[1] - g;   if (a<0) a = -a;
        dist += a;
        a = n[2] - b;   if (a<0) a = -a;
        dist += a;
        if (dist<bestd) { bestd = dist; bestpos = i; }
        biasdist = dist - ((*p) >> 12);
        if (biasdist<bestbiasd) { bestbiasd = biasdist; bestbiaspos = i; }
        betafreq = (*f >> 10);
        *f++ -= betafreq;
        *p++ += (betafreq << 10);
    }
    freq[bestpos] += 64;
    bias[bestpos] -= 65536;
    return(bestbiaspos);
}




/* Move adjacent neurons by precomputed alpha*(1-((i-j)^2/[r]^2)) in radpower[|i-j|]
--------------------------------------------------------------------------------- */

static void alterneigh(long rad, long i, long r, long g, long b)
{
    long j, k, lo, hi, a;
    long *p, *q;

    lo = i - rad;   if (lo<-1) lo = -1;
    hi = i + rad;   if (hi>256) hi = 256;

    j = i + 1;
    k = i - 1;
    q = radpower;
    while ((j<hi) || (k>lo)) {
        a = (*(++q));
        if (j<hi) {
            p = network[j];
            *p -= (a*(*p - r)) / 0x40000;
            p++;
            *p -= (a*(*p - g)) / 0x40000;
            p++;
            *p -= (a*(*p - b)) / 0x40000;
            j++;
        }
        if (k>lo) {
            p = network[k];
            *p -= (a*(*p - r)) / 0x40000;
            p++;
            *p -= (a*(*p - g)) / 0x40000;
            p++;
            *p -= (a*(*p - b)) / 0x40000;
            k--;
        }
    }
}
static void inxbuild()
{
    register long i, j, smallpos, smallval;
    register long *p, *q;
    int previouscol, startpos;

    previouscol = 0;
    startpos = 0;
    for (i = 0; i<256; i++) {
        p = network[i];
        smallpos = i;
        smallval = p[1];			/* index on g */
                                    /* find smallest in i..netsize-1 */
        for (j = i + 1; j<256; j++) {
            q = network[j];
            if (q[1] < smallval) {		/* index on g */
                smallpos = j;
                smallval = q[1];	/* index on g */
            }
        }
        q = network[smallpos];
        /* swap p (i) and q (smallpos) entries */
        if (i != smallpos) {
            j = q[0];   q[0] = p[0];   p[0] = j;
            j = q[1];   q[1] = p[1];   p[1] = j;
            j = q[2];   q[2] = p[2];   p[2] = j;
            j = q[3];   q[3] = p[3];   p[3] = j;
        }
        /* smallval entry is now in position i */
        if (smallval != previouscol) {
            netindex[previouscol] = (startpos + i) >> 1;
            for (j = previouscol + 1; j<smallval; j++) netindex[j] = i;
            previouscol = smallval;
            startpos = i;
        }
    }
    netindex[previouscol] = (startpos + 255) >> 1;
    for (j = previouscol + 1; j<256; j++) netindex[j] = 255; /* really 256 */
}

/* Main Learning Loop
------------------ */

void neuquantcq(unsigned char *thepic, unsigned long pixels, unsigned char* pal, unsigned long colorDepth,unsigned long cycles)
{
    long  j, b, g, r;
    long radius, rad, alpha, delta;
    unsigned char *p;
    unsigned char *lim;
    unsigned long i,step;

    long *pp;


    unsigned long lengthcount = pixels*colorDepth;

    for (i = 0; i<256; i++) {
        pp = network[i];
        pp[0] = pp[1] = pp[2] = i << 4;
        freq[i] = 256;	/* 1/256 */
        bias[i] = 0;
    }



    p = thepic;
    lim = thepic + lengthcount;

    delta = pixels / cycles;
    alpha = 1024;
    radius = 2048;

    rad = 32;
    for (i = 0; i<32; i++)
        radpower[i] = 1024 * ((1024 - i*i) / 4);


    if ((lengthcount % 499) != 0) step = 4 * 499;
    else {
        if ((lengthcount % 491) != 0) step = 4 * 491;
        else {
            if ((lengthcount % 487) != 0) step = 4 * 487;
            else step = 4 * 503;
        }
    }

    i = 0;
    while (i < pixels) {
        r = p[0] << 4;
        g = p[1] << 4;
        b = p[2] << 4;
        j = contest(r, g, b);

        long *n;
        n = network[j];				/* alter hit neuron */
        *n -= (alpha*(*n - r)) / 1024;
        n++;
        *n -= (alpha*(*n - g)) / 1024;
        n++;
        *n -= (alpha*(*n - b)) / 1024;

        if (rad) alterneigh(rad, j, r, g, b);   /* alter neighbours */

        p += step;
        if (p >= lim) p -= lengthcount;

        i++;
        if (i%delta == 0) {
            alpha -= alpha / 30;
            radius -= radius / 30;
            rad = radius >> 6;
            if (rad <= 1) rad = 0;
            for (j = 0; j<rad; j++)
                radpower[j] = alpha*(((rad*rad - j*j) * 256) / (rad*rad));
        }
    }
    //Unbias network to give byte values 0..255 and record position i to prepare for sort
    for (i = 0; i<256; i++) {
        r = network[i][0] + 8;
        g = network[i][1] + 8;
        b = network[i][2] + 8;
        r >>= 4;
        g >>= 4;
        b >>= 4;
        if (r>255)r = 255;
        if (g>255)g = 255;
        if (b>255)b = 255;
        *pal++ = (unsigned char)r;
        *pal++ = (unsigned char)g;
        *pal++ = (unsigned char)b;
        *pal++ = 0xff;
        //network[i][3] = i;			/* record colour no */
    }



}
